// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file AbstractFunction.h */

#ifndef AMDIS_ABSTRACTFUNCTION_H
#define AMDIS_ABSTRACTFUNCTION_H

#include "Global.h"
#include "MemoryManager.h"

namespace AMDiS {

  // ============================================================================
  // ===== class AbstractFunction ===============================================
  // ============================================================================

  /** 
   * \ingroup Common
   * 
   * \brief
   * An AbstractFunction object represents a function 
   * f : ArgumentType -> ReturnType. 
   *
   * AbstractFunction is a pure virtual class interface class.
   * To create your own function you have to derive AbstractFunction and
   * overload operator(). 
   */
  template<typename ReturnType, typename ArgumentType>
  class AbstractFunction
  {
  public:
    /** \brief
     * Constructor.
     */
    AbstractFunction(int degree = 0) : 
      degree_(degree) 
    {};

    virtual ~AbstractFunction() {};

    /** \brief
     * Returns \ref degree_.
     */
    inline int getDegree() const { 
      return degree_; 
    };

    /** \brief
     * Deligates the evaluation to overriden method f.
     */
    virtual ReturnType operator()(const ArgumentType& x) const = 0;

  protected:
    int degree_;
  };

  // ============================================================================
  // ===== class BinaryAbstractFunction =========================================
  // ============================================================================

  /**
   * \ingroup Common
   *
   * \brief
   * Interface for binary functions. 
   */
  template<typename ReturnType, 
	   typename ArgumentType1, 
	   typename ArgumentType2>
  class BinaryAbstractFunction
  {
  public:
    /** \brief
     * Constructor.
     */
    BinaryAbstractFunction(int degree = 0) : 
      degree_(degree) 
    {};

    virtual ~BinaryAbstractFunction() {};

    /** \brief
     * Returns \ref degree_.
     */
    inline int getDegree() const { 
      return degree_; 
    };

    /** \brief
     * Deligates the evaluation to overriden method f.
     */
    virtual ReturnType operator()(const ArgumentType1& x,
				  const ArgumentType2& y) const = 0;

  protected:
    int degree_;
  };

  // ============================================================================
  // ===== class TertiaryAbstractFunction =======================================
  // ============================================================================

  /**
   * \ingroup Common
   *
   * \brief
   * Interface for tertiary functions. 
   */
  template<typename ReturnType, 
	   typename ArgumentType1, 
	   typename ArgumentType2,
	   typename ArgumentType3>
  class TertiaryAbstractFunction
  {
  public:
    /** \brief
     * Constructor.
     */
    TertiaryAbstractFunction(int degree = 0) : 
      degree_(degree) 
    {};

    virtual ~TertiaryAbstractFunction() {};

    /** \brief
     * Returns \ref degree_.
     */
    inline int getDegree() const { 
      return degree_; 
    };

    /** \brief
     * function evaluation.
     */
    virtual ReturnType operator()(const ArgumentType1& x,
				  const ArgumentType2& y,
				  const ArgumentType3& z) const = 0;

  protected:
    int degree_;
  };


  // ============================================================================
  // ===== class QuartAbstractFunction =======================================
  // ============================================================================

  /**
   * \ingroup Common
   *
   * \brief
   * Interface for quart functions. 
   */
  template<typename ReturnType, 
	   typename ArgumentType1, 
	   typename ArgumentType2, 
	   typename ArgumentType3,
	   typename ArgumentType4>
  class QuartAbstractFunction
  {
  public:
    /** \brief
     * Constructor.
     */
    QuartAbstractFunction(int degree = 0) : 
      degree_(degree) 
    {};

    virtual ~QuartAbstractFunction() {};

    /** \brief
     * Returns \ref degree_.
     */
    inline int getDegree() const { 
      return degree_; 
    };

    /** \brief
     * function evaluation.
     */
    virtual ReturnType operator()(const ArgumentType1& x,
				  const ArgumentType2& y,
				  const ArgumentType3& z,
				  const ArgumentType4& u) const = 0;

  protected:
    int degree_;
  };




}

#endif // AMDIS_ABSTRACTFUNCTION_H
